@@ -1,5 +1,5 @@
-from .shapes import Field, StateLayout, StateShapeError
-from .shapes import count_qubits, real_imag, real_imag_tensor
+from .shapes import StateLayout, StateShapeError
+from .shapes import count_qubits
 from .shapes import count_qubits_tensor
 from .shapes import count_qubits_gate_matrix
 from .shapes import hilbert_space_dim
@@ -8,4 +8,4 @@ from .shapes import subset_roll_to_back, subset_roll_to_front
 from .conversions import to_tensor_layout, to_vector_layout
 from .innerprod import inner_product, norm_squared
 from .innerprod import abs_squared
-from .innerprod import expectation_value
+from .innerprod import diag_expectation_value
